from .base import *

class Item(Object):
    listing_endpoint="/items"
    endpoint="/items/{guid}"

    @property
    @lazy
    def bom(self):
        return self._client.BOM(self.guid)

    @property
    @lazy
    def category(self):
        return self._client.ItemCategory(self.__dict__["category"]["guid"])

    @property
    @lazy
    def revisions(self):
        return self._client.Listing(self._client.ItemRevision, endpoint=f"/items/{self.guid}/revisions")

    @property
    @lazy
    def quality_processes(self):
        return self._client.Listing(self._client.QualityProcess, endpoint=f"/items/{self.guid}/quality")

    @property
    @lazy
    def where_used(self):
        return self._client.Listing(self._client.Item, endpoint=f"/items/{self.guid}/whereused")

    @property
    @lazy
    def file_associations(self):
        assocs = self._client.Listing(self._client.ItemFileAssociation, endpoint=f"/items/{self.guid}/files")
        for assoc in assocs:
            assoc.item=self
        return assocs

    @property
    @lazy
    def thumbnail(self):

        return self._client._get(f"/items/{self.guid}/image/content")

    @property
    def working_revision(self):
        rev = [x for x in self.revisions if x.status==0]
        if rev:
            return rev[0]
        else:
            return None

    @property
    def effective_revision(self):
        rev = [x for x in self.revisions if x.status==1]
        if rev:
            return rev[0]
        else:
            return None

    @property
    @lazy
    def lifecycle_phase(self):

        return self._client.ItemLifecyclePhase(self.lifecyclePhase['guid'])

class ItemLifecyclePhase(Object):
    listing_endpoint = "/settings/items/lifecyclephases"
    

class ItemCategory(Object):

    listing_endpoint="/settings/items/categories"
    endpoint=listing_endpoint+"/{guid}"
    _can_paginate=False

    @property
    @lazy
    def attributes(self):
        return self._client.Listing(self._client.ItemCategoryAttribute, endpoint=f"/settings/items/categories/{self.guid}/attributes")

class ItemCategoryAttribute(Object):
    pass

class ItemFileAssociation(Object):
    @property
    @lazy
    def file(self):
        return self._client.File(self.__dict__['file']['guid'])

    @property
    def endpoint(self):
        return f"/items/{self.change.guid}/files/{self.guid}"

class ItemRevision(Object):
    pass


from .base import *

class File(Object):
    listing_endpoint="/files"
    endpoint="/files/{guid}"

    @property
    @lazy
    def author(self):
        return self._client.User(self.__dict__['author'])

    def update_file_contents(self, filename):

        """
        Administratively replace this File's document with a new one.

        Required arguments:
        - filename   - Path to the file to upload
        """

        with open(filename, "r+") as file:
            self._client._post(
                f"{self.endpoint}/content",
                files={
                    'content':file
                    }
                )

    def check_out(self, comment="Checking out for review"):
        """
        Checks out the File for edits
        """

        data={
            "checkedOut":True,
            "comment":comment,
            "file":{
                "guid":self.guid
            }
        }
        return self._client._post(
            "/files/checkoutstatuschanges",
            data=data
            )

    def cancel_check_out(self, comment="Cancelling file checkout"):

        """
        Reverts File to "Checked In" state.
        """

        data={
            "checkedOut":False,
            "comment":comment,
            "file":{
                "guid":self.guid
            }
        }
        return self._client._post(
            "/files/checkoutstatuschanges",
            data=data
            )

    def check_in(self, filename, comment="Checking in for update"):

        """
        Upload and check in new version of file to replace existing version.
        """

        with open(filename, "r+") as file:

            files = {
                "file.content":file
            }

            data = {
                "file.guid": self.guid,
                "newEdition":True
            }

            returnval= self._client._post(
                "/files/checkoutstatuschanges",
                data=data,
                files=files
                )

        return returnval

    @property
    @lazy
    def corrections(self):
        return self._client.Listing(self._client.FileCorrection, endpoint=f"/files/{self.guid}/corrections")

    @property
    @lazy
    def editions(self):
        return self._client.Listing(self._client.File, endpoint=f"/files/{self.guid}/editions")

    @property
    @lazy
    def content(self):
        return self._client._get(f"/files/{self.guid}/content").content




class FileCorrection(Object):
    pass


class FileCategory(Object):

    listing_endpoint="/settings/files/categories"
    endpoint=listing_endpoint+"/{guid}"
    _can_paginate=False

class FileAttribute(Object):

    listing_endpoint="/settings/files/attributes"


from .base import *

class Ticket(Object):
    listing_endpoint = "/tickets"
    endpoint=listing_endpoint+"/{guid}"

    def set_status(self, status):

        """
        Change the Ticket status.

        Required argument:

        - status - one of `NOT_STARTED`, `IN_PROGRESS`, or `COMPLETE`
        """

        if status not in ['NOT_STARTED', 'IN_PROGRESS', 'COMPLETE']:

            raise ValueError("`status` must be one of `NOT_STARTED`, `IN_PROGRESS`, or `COMPLETE`.")

        body={
            "ticket": {
                "guid":self.guid
            },
            "status": status
        }

        x=self._client._post("/tickets", data=body)

        self.__dict__.update(x.json())


    @property
    @lazy
    def file_associations(self):
        assocs = self._client.Listing(self._client.TicketFileAssociation, endpoint=f"/tickets/{self.guid}/files")
        for assoc in assocs:
            assoc.ticket=self
        return assocs

    @property
    @lazy
    def change_associations(self):
        assocs = self._client.Listing(self._client.TicketChangeAssociation, endpoint=f"/tickets/{self.guid}/changes")
        for assoc in assocs:
            assoc.ticket=self
        return assocs

    @property
    @lazy
    def quality_associations(self):
        assocs = self._client.Listing(self._client.TicketQualityAssociation, endpoint=f"/tickets/{self.guid}/quality")
        for assoc in assocs:
            assoc.ticket=self
        return assocs

    @property
    @lazy
    def ticket_associations(self):
        assocs = self._client.Listing(self._client.TicketTicketAssociation, endpoint=f"/tickets/{self.guid}/tickets")
        for assoc in assocs:
            assoc.ticket=self
        return assocs


    def add_file(self, file_obj):

        """
        Add File connection to Change.

        Args:

        * file_obj - A File object
        """

        data={
            "file":{
                "guid":file_obj.guid
            }
        }
        x=self._client._post(f"/tickets/{self.guid}/files", data=data)

        return self._client.TicketFileAssociation(x.json())

class TicketFileAssociation(Object):
    
    @property
    @lazy
    def file(self):
        return self._client.File(self.__dict__['file']['guid'])

    @property
    def endpoint(self):
        return f"/tickets/{self.ticket.guid}/files/{self.guid}"

class TicketChangeAssociation(Object):
    @property
    @lazy
    def change(self):
        return self._client.Change(self.__dict__['change']['guid'])

    @property
    def endpoint(self):
        return f"/tickets/{self.ticket.guid}/changes/{self.guid}"

class TicketQualityAssociation(Object):
    
    @property
    @lazy
    def quality(self):
        return self._client.QualityProcess(self.__dict__['quality']['guid'])

    @property
    def endpoint(self):
        return f"/tickets/{self.ticket.guid}/quality/{self.guid}"

class TicketTicketAssociation(Object):
    
    @property
    @lazy
    def other_ticket(self):
        return self._client.Ticket(self.__dict__['ticket']['guid'])

    @property
    def endpoint(self):
        return f"/tickets/{self.ticket.guid}/quality/{self.guid}"

class TicketAttribute(Object):

    listing_endpoint="/settings/tickets/attributes"


from .base import *

class Integration(Object):

    listing_endpoint="/outboundintegrations"
    endpoint=listing_endpoint+"/{guid}"

    @property
    @lazy
    def administrators(self):
        data = self._client._get("/outboundintegrations/{self.guid}/administrators")
        return [self._client.User(guid=x) for x in [entry["guid"] for entry in data["results"]]]

    @property
    @lazy
    def events(self):
        events = self._client.Listing(self._client.Event, endpoint=f"/outboundintegrations/{self.guid}/events")
        for event in events:
            event.integration = self
        return events


class Event(Object):

    @property
    def endpoint(self):
        return self.integration.endpoint+f"/events/{self.guid}"

    @property
    @lazy
    def items(self):
        items = self._client.Listing(self._client.EventItem, endpoint=f"{self.endpoint}/items")
        for item in items:
            item.event = self
        return items

    def reconcile(self):
        """
        Mark all Items associated with this event as reconciled.
        """

        return self._client._put(self.endpoint, data={"itemsReconciled":True})

    def unreconcile(self):
        """
        Mark all Items associated with this event as not reconciled.
        """

        return self._client._put(self.endpoint, data={"itemsReconciled":False})


    @property
    @lazy
    def change(self):
        return self._client.Change(guid=self.change["guid"])

    def item_ids(self, reconciled="any"):

        """
        Get IDs of Items associated with this Event.

        Kwargs:

        * reconciled - one of `"true"`, `"false"`, or `"any"`
        """
        
        if reconciled not in ["true", "false", "any"]:
            raise ValueError('`reconciled` must be one of `"true"`, `"false"`, or `"any"`')

        url=f"{self.endpoint}/itemguids"

        data=self._client._get(url, params={"reconciled":reconciled})

        return data["results"]



class EventItem(Object):

    @property
    def endpoint(self):
        return f"/outboundintegrations/{self.event.integration.guid}/events/{self.event.guid}/items/{self.guid}"

    def reconcile(self):
        """
        Mark this event item as reconciled.
        """

        return self._client._put(self.endpoint, data={"reconciled":True})

    def unreconcile(self):
        """
        Mark this event item as not reconciled.
        """

        return self._client._put(self.endpoint, data={"reconciled":False})

    @property
    @lazy
    def item(self):

        return self._client.Item(guid=self.effectiveItemRevision["guid"])

class Trigger(Object):

    listing_endpoint="/settings/integrations/triggers"
    endpoint=listing_endpoint+"/{guid}"



import requests
import sys
import time
import mistletoe
from collections import deque
import webbrowser
import inspect
import warnings
import urllib3
import ssl

import pyrena.classes as classes

class ArenaHTTPError(Exception):
    pass

def disable (f):
    def wrapper(*args, **kwargs):
        self=args[0]
        endpoint=args[1]

        if endpoint not in ["/login","/logout"]:
            raise AttributeError(f"Function {self.__class__.__name__}.{f.__name__} is currently disabled")

        return f(*args, **kwargs)

    wrapper.__name__=f.__name__
    return wrapper

class Arena():

    def __init__(self, username=None, password=None, base_url="https://api.arenasolutions.com/v1", ssl_verify=True, user_agent="Python Requests", verbose=False):

        """
        Creates the Arena client

        Optional arguments:
        - base_url      - change the base URL. Set to `https://api.arenagov.com/v1/v1` for government users.
        - ssl_verify    - Use strict SSL verification in requests. May need to be set to False when accessing Arena from a corporate-controlled network.
        - user_agent    - Set a User-Agent header.
        - verbose       - Verbosely output API requests to console
        """

        #set properties
        self.base_url = base_url
        self.ssl_verify = ssl_verify

        #if ssl_verify is false, disable insecure http warnings
        if not ssl_verify:
            warnings.filterwarnings("ignore", category=urllib3.exceptions.InsecureRequestWarning)

        self.user_agent = user_agent

        self._debug=verbose

        self.__dict__['_username']=username
        self.__dict__['_password']=password


        self.headers={
            "User-Agent": self.user_agent,
            "Content-Type": "application/json"
        }

        self._reqs_remaining=25000

        #login
        try:
            self.login()
        except ArenaHTTPError:
            print("Unable to log in. Call `login(username, password)` on this client to authenticate with Arena.")
        except ssl.SSLCertVerificationError:
            print("Unable to establish a secure connection to Arena. Use a different internet connection or use `ssl_verify=False` to allow insecure requests.")

        #initialize cache
        self._cache={}

    def __getattr__(self, name):

        if name in ['username', '_username', 'password', '_password']:
            raise AttributeError

        try:
            obj= getattr(classes, name)

            if not isinstance(obj, type):
                raise(KeyError)
        except:
            try:
                return self.__dict__[name]
            except KeyError:
                raise AttributeError

        obj._client=self

        return obj

    def __repr__(self):
        return f"<Arena(user={self.__dict__['_username']})>"

    def login(self, username=None, password=None):

        """
        Log into Arena

        Optional arguments:
        - username    - Username. Defaults to username given at client creation if not specified.
        - password    - Password. Defaults to password given at client creation if not specified.
        """

        self.__dict__["_username"] = username or self.__dict__["_username"]
        self.__dict__["_password"] = password or self.__dict__["_password"]

        login_data = self._post(
            "/login",
            data={
                "email":self.__dict__["_username"],
                "password":self.__dict__["_password"]
                }
            )


        #set auth token
        self.headers["arena_session_id"]=login_data["arenaSessionId"]

        #set workspace id and name
        self._workspace_id=login_data["workspaceId"]
        self._workspace_name=login_data["workspaceName"]

        #set time to re-auth, 1 min before expiry
        self.reauth_utc = int(time.time())+60*60*80

        if self._debug:
            print(login_data)

    def logout(self):

        """
        Log out of Arena.
        """

        self._put("/logout")
        self.__dict__["_username"]=None
        self.__dict__["_password"]=None

    def _fetch(self, method, endpoint, params={}, data={}, files={}, headers={}, **kwargs):

        #reauth 1 min before expiry
        if endpoint != "/login" and int(time.time())> self.reauth_utc:
            self.login()

        url = f"{self.base_url}{'' if endpoint.startswith('/') else '/'}{endpoint}"

        req_headers=self.headers
        for header in headers:
            req_headers[header]=headers[header]

        try:
            req = getattr(requests, method.lower())(
                url, 
                headers=req_headers, 
                params=params,
                json=data, 
                files=files,
                verify=self.ssl_verify,
                )
        except requests.exceptions.SSLError:
            print("Unable to establish a secure connection to Arena. Use a different internet connection or use `ssl_verify=False` to allow insecure requests.")
            return

        if self._debug:
            print(f"{method.upper()} {endpoint} - {req.status_code}")

        if req.status_code >=400:
            data=req.json()
            raise ArenaHTTPError(f"HTTP Error {req.status_code} - {data['errors'][0]['message']} (Arena Error Code {data['errors'][0]['code']})")

        req.raise_for_status()


        #update reqs remaining
        self._reqs_remaining = req.headers.get("X-Arena-Requests-Remaining", self._reqs_remaining)

        return req.json()

    def _get(self, endpoint, params={}, **kwargs):
        return self._fetch("get", endpoint, params=params, **kwargs)

    #@disable
    def _post(self, endpoint, data={}, files={}, **kwargs):
        return self._fetch("post", endpoint, data=data, files=files, **kwargs)

    #@disable
    def _put(self, endpoint, data={}, **kwargs):
        return self._fetch("put", endpoint, data=data, **kwargs)

    #@disable
    def _delete(self, endpoint, data={}, **kwargs):
        return self._fetch("delete", endpoint, data=data, **kwargs)

    def Listing(self, obj, endpoint=None, limit=400, offset=0,  **kwargs):

        '''
        Obtains a listing of objects

        Required arguments:
        - obj       - Object class to aquire. For example, client.QualityProcess

        Optional arguments:
        - endpoint  - The endpoint to use. Defaults to obj.listing_endpoint property
        - limit     - The maximum number to fetch. Defaults to 400, the maximum number of objects that may be returned by one Arena API call.
        - offset    - Offset for result returns. Defaults to 0 (no offset). Mainly used in recursive pagination calls.
        '''

        if isinstance(obj, str):
            obj=getattr(classes, obj)

        endpoint = endpoint or obj.__dict__.get("listing_endpoint")
        if not endpoint:
            raise AttributeError(f"Object type `{obj.__name__}` cannot be globally listed")

        #assemble search params if needed
        params={}
        for entry in kwargs:
            if isinstance(kwargs[entry], classes.Object):
                params[f"{entry}.guid"] = kwargs[entry].guid
            elif isinstance(entry, str):
                params[entry]=kwargs[entry]
            else:
                params[entry.guid]=kwargs[entry]

        if getattr(obj, "_can_paginate", True):

            params["limit"]=min(limit, 400) if limit else 400
            params["offset"]=offset  

        data=self._get(endpoint, params=params)

        listing=[obj(**dictionary) for dictionary in data["results"]]

        for item in listing:
            item._client=self


        if len(listing)==limit:
            listing+=self.Listing(
                    obj, 
                    endpoint, 
                    limit=400 if limit else None,
                    offset=offset+400,
                    **kwargs
                    )
        
        return listing



    def Stream(self, obj, endpoint=None, limit=400, offset=0, delay=60, **kwargs):

        """
        Similar to Listing, but yields live results without end. Does not yield pre-existing content.

        Required arguments:
        - obj       - Object class to aquire. For example, client.QualityProcess

        Optional arguments:
        - endpoint  - The endpoint to use. Defaults to obj.listing_endpoint property
        - limit     - The maximum number to fetch. Defaults to 400, the maximum number of objects that may be returned by one Arena API call.
        - offset    - Offset for result returns. Defaults to 0 (no offset). Mainly used in recursive pagination calls.
        - delay     - Time to sleep (in seconds) between API calls.
        """

        initial = self.Listing(obj, endpoint, limit, offset, **kwargs)

        already_seen=deque([], 1000)
        for item in initial:
            already_seen.append(item.__repr__())

        while True:

            time.sleep(delay)

            for entry in self.Listing(*args, **kwargs):

                if entry.__repr__ in already_seen:
                    continue

                already_seen.append(entry.__repr__())
                yield entry

    def clear_cache(self, item=None):

        """
        Dumps cached versions of an item, causing it to be reloaded from Arena on next access.

        Optional arguments:

        - item    - Any object corresponding to any Arena record. If None, dumps the entire cache.
        """

        if item and isinstance(item, classes.Object):
            item=item.cache_key

        if item:
            classes.OBJ_CACHE.pop(item, None)

        else:
            classes.OBJ_CACHE={}

        return True


def docs():

    """
    Generate and display this documentation page.
    """

    classlist = {name:cls for name, cls in list(classes.__dict__.items()) if isinstance(cls, type) and name!="Object"}
    classlist["Arena"]=Arena

    #page header
    md = "# `Pyrena Docs`\n\n"

    #Iterate through classes (including the Arena client class)
    for class_name in sorted(classlist.keys()):
        class_obj = classlist[class_name]

        methods=[x for x in sorted(dir(class_obj)) if callable(getattr(class_obj, x)) and not x.startswith('_')]

        params=inspect.signature(class_obj.__init__).parameters
        class_args=[f"{'*' if params[x].name=='args' else ''}{'**' if params[x].name=='kwargs' else ''}{params[x].name}{f'={params[x].default.__repr__()}' if params[x].default != params[x].empty else ''}" for x in params]

        md+= f"## `{class_name}({', '.join(class_args)})`\n\n{f'{class_obj.__init__.__doc__}' if class_obj.__init__.__doc__ else ''}\n\n"

        #iterate through callable methods of that class
        for method in methods:

            params=inspect.signature(getattr(class_obj, method)).parameters

            args = [f"{'*' if params[x].name=='args' else ''}{'**' if params[x].name=='kwargs' else ''}{params[x].name}{f'={params[x].default.__repr__()}' if params[x].default != params[x].empty else ''}" for x in params if params[x].name!='self']


            md+= f"### `{method}({', '.join(args)})`\n\n{getattr(class_obj, method).__doc__}\n\n"

    with open("pyrena.html", "w+") as file:
        file.write(mistletoe.markdown(md))
        file.truncate()

    webbrowser.open("pyrena.html")


from .base import *

class Change(Object):
    listing_endpoint="/changes"
    endpoint="/changes/{guid}"

    @property
    @lazy
    def file_associations(self):
        assocs = self._client.Listing(self._client.ChangeFileAssociation, endpoint=f"/changes/{self.guid}/files")
        for assoc in assocs:
            assoc.change=self
        return assocs

    def add_file(self, file_obj):

        """
        Add File connection to Change.

        Args:

        * file_obj - A File object
        """

        data={
            "file":{
                "guid":file_obj.guid
            }
        }
        x=self._client._post(f"/changes/{self.guid}/files", data=data)

        return self._client.ChangeFileAssociation(x.json())

    @property
    @lazy
    def item_associations(self):
        assocs = self._client.Listing(self._client.ChangeItemAssociation, endpoint=f"/changes/{self.guid}/items")
        for assoc in assocs:
            assoc.change=self
        return assocs

    @property
    @lazy
    def implementation_file_associations(self):

        files = self._client.Listing(self._client.ImplementationFileAssociation, endpoint=f"/changes/{self.guid}/implementationfiles")
        for file in files:
            file.change=self
        return files


    def add_item(self, item_obj, lifecycle_phase=None, new_rev=None):

        """
        Add Item revision to Change.

        Args:

        * item_obj - An Item object. The working revision will be selected.

        Kwargs:

        * lifecycle_phase - An ItemLifecyclePhase object. Defaults to `item_obj.lifecycle_phase` if not specified
        """

        if not lifecycle_phase:
            lifecycle_phase = item_obj.lifecycle_phase

        if lifecycle_phase.stage=='PRELIMINARY':
            raise ValueError("Must specify a DESIGN or PRODUCTION lifecycle phase")

        data={
            "filesView":{
                "includedInThisChange":True,
                "notes":f"Add {item_obj.name}"
            },
            "sourcingView":{
                "includedInThisChange":True,
                "notes":f"Add {item_obj.name}"
            },
            "specsView":{
                "includedInThisChange":True,
                "notes":f"Add {item_obj.name}"
            },
            "bomView":{
                "includedInThisChange":True,
                "notes":f"Add {item_obj.name}"
            },
            "newRevisionNumber": new_rev,
            "newItemRevision":{
                "guid":item_obj.working_revision.guid
            },
            "newLifecyclePhase":{
                "guid": lifecycle_phase.guid
            }
        }
        self._client._post(f"/changes/{self.guid}/items", data=data)

class ChangeFileAssociation(Object):
    
    @property
    @lazy
    def file(self):
        return self._client.File(self.__dict__['file']['guid'])

    @property
    def endpoint(self):
        return f"/changes/{self.change.guid}/files/{self.guid}"

class ChangeItemAssociation(Object):
    
    @property
    @lazy
    def item(self):
        return self._client.Item(self.__dict__['item']['guid'])

    @property
    def endpoint(self):
        return f"/changes/{self.change.guid}/files/{self.guid}"


class ImplementationFileAssociation(Object):

    @property
    def endpoint(self):
        return f"/changes/{self.change.guid}/files/{self.guid}"

    @property
    @lazy
    def file(self):
        return self._client.File(self.__dict__['file']['guid'])
    
class ChangeCategory(Object):

    listing_endpoint="/settings/changes/categories"
    endpoint=listing_endpoint+"/{guid}"
    _can_paginate=False

class ChangeAttribute(Object):

    listing_endpoint="/settings/changes/attributes"



#from .base     import *

#These classes all subclass .base.Object
from .bom        import *
from .change     import *
from .eventqueue import *
from .extract    import *
from .file       import *
from .item       import *
from .quality    import *
from .supplier   import *
from .tickets    import *
from .user       import *


from .base import *

class QualityProcess(Object):
    listing_endpoint = "/qualityprocesses"
    endpoint=listing_endpoint+"/{guid}"

    @property
    @lazy
    def template(self):
        return self._client.QualityProcessTemplate(self.__dict__["template"]["guid"])

    @property
    @lazy
    def steps(self):
        steps = self._client.Listing(self._client.QualityProcessStep, endpoint=f"/qualityprocesses/{self.guid}/steps")
        for step in steps:
            step.quality_process = self
        return steps

    @property
    @lazy
    def owner(self):
        return self._client.User(self.__dict__['owner']['guid'])

class QualityProcessNumberFormat(Object):
    listing_endpoint = "/settings/qualityprocesses/numberformats"
    endpoint=listing_endpoint+"/{guid}"

class QualityProcessStep(Object):
  
    @property
    def endpoint(self):
        return f"/qualityprocess/{self.quality_process.guid}/steps/{self.guid}"

    @property
    @lazy
    def assignees(self):
        return [self._client.User(guid) for guid in [x['guid'] for x in self.__dict__['assignees']['users']]]

    def complete(self, comment=""):

        """
        Mark the Step as Complete.
        """

        data={  
           "Quality Process":{
               "guid":self.qualityprocess.guid,
               "step": {
                   "guid": self.guid
               }
           },
           "complete":True,
           "comment":comment
        }

        return self._client._post(
            f"/qualityprocess/statuschanges",
            data=data
            )
          
    def reopen(self, comment=""):

        """
        Reopen a completed Step.
        """

        data={  
           "Quality Process":{
               "guid":self.qualityprocess.guid,
               "step": {
                   "guid": self.guid
               }
           },
           "complete":False,
           "comment":comment
        }

        return self._client._post(
            f"/qualityprocess/statuschanges",
            data=data
            )  

class QualityProcessTemplate(Object):
    listing_endpoint="/settings/qualityprocesses/templates"
    endpoint = listing_endpoint+"/{guid}"

    @property
    @lazy
    def number_formats(self):
        return [self._client.QualityProcessNumberFormat(**dataset) for dataset in  self.__dict__.get("numberFormats", [])]

    @property
    @lazy
    def default_number_format(self):
        return self._client.QualityProcessNumberFormat(self.__dict__['defaultNumberFormat']['guid'])

    @property
    @lazy
    def steps(self):
        steps=[self._client.QualityProcessTemplateStep(**dataset) for dataset in self.steps]
        for step in steps:
            step.template=self

class QualityProcessTemplateStep(Object):
    pass

class QualityAttribute(Object):

    listing_endpoint="/settings/quality/attributes"


from .base import *

class User(Object):
    listing_endpoint = "/settings/users"
    endpoint=listing_endpoint+"/{guid}"

    @property
    def name(self):
        return self.fullName


class UserGroup(Object):
    listing_endpoint = "/settings/usergroups"
    endpoint=listing_endpoint+"/{guid}"

    @property
    @lazy
    def users(self):

        return self._client.Listing(self._client.User, endpoint=f"/settings/usergroups/{self.guid}/users")



from .base import *

class BOM(Object):
    endpoint="/items/{guid}/bom"

    @property
    @lazy
    def settings(self):
        return self._client.get(f"/items{self.guid}/bom/settings")

    @property
    @lazy
    def lines(self):

        lines=[self._client.BOMLine(**dataset) for dataset in self.results]
        for line in lines:
            line.bom=self


class BOMLine(Object):

    @property
    @lazy
    def endpoint(self):
        return f"/items/{self.bom.guid}/bom/{self.guid}"
    
    @property
    @lazy
    def item(self):
        return self._client.Item(self.__dict__["item"]["guid"])


import time

def lazy(f):

    def wrapper(self):

        if f.__name__ in self._cache:
            return self._cache[f.__name__]
        else:

            return_value = f(self)
            self._cache[f.__name__]=return_value
            return return_value

    wrapper.__name__=f.__name__
    return wrapper



OBJ_CACHE={}

class Object():

    def __new__(cls, guid=None, **kwargs):

        cache_key = f"{cls.__name__};{guid}"

        if not kwargs:
            obj=OBJ_CACHE.get(cache_key)
            if obj:
                return obj

        obj = super(Object, cls).__new__(cls)

        return obj

    def __init__(self, guid=None, **kwargs):

        """
        Use one of the following:

        guid - object ID to fetch from Arena

        or

        **kwargs - Dictionary of properties to set on a blank object instance that does not yet exist in Arena. Call `create()` once all desired properties are set.
        """

        #skip if cached
        cache_key = f"{self.__class__.__name__};{guid}"

        if guid and not kwargs:
            if cache_key in OBJ_CACHE:
                return

            self.endpoint=self.__class__.__dict__.get('endpoint')
            if not self.endpoint:
                raise ValueError(f"{self.__class__.__name__} objects cannot be directly obtained by GUID.")

            if not isinstance(self.endpoint, property):
                self.endpoint=self.endpoint.format(guid=guid)
            
            data = self._client._get(self.endpoint)

            self.__dict__.update(data)

        elif kwargs:
            self.guid=guid
            self.__dict__.update(kwargs)
            if not isinstance(self.__class__.__dict__.get('endpoint'), property):
                self.endpoint=self.__class__.__dict__.get('endpoint').format(guid=self.guid)

        else:
            raise AttributeError("Must specify either `guid` or `**kwargs`")

        self._cache={}

        OBJ_CACHE[cache_key]=self

    def __repr__(self):

        if self._client._debug:

            return f"<{self.__class__.__name__}(id={self.guid})>"

        else:
            if self.__dict__.get("number"):
                return f"<{self.__class__.__name__}(number={self.number})>"
            else:
                return f"<{self.__class__.__name__}(name={self.name})>"

    def __str__(self):
        if "name" in self.__dict__:
            return f"<{self.__class__.__name__}(name={self.name})>"
        else:
            return self.__repr__

    def __eq__(self, other):
        return self.__repr__==other.__repr__

    # def __getattr__(self, name):

    #     # "2018-06-01T06:59:59Z"
    #     if name.endswith("_utc"):
    #         target_name=f"{name.split('_utc')[0]}DateTime"
    #         target_time=self.__dict__[target_name]
    #         return int(time.mktime(time.strptime(self.__dict__[target_name], "%Y-%m-%dT%H:%M:%SZ")))

    #     return self.__dict__[name]



    def refresh(self, value=None):
        """
        Dumps cache for lazily loaded object properties. Dumped properties will be reloaded from Arena at next access.

        Optional arguments:

        value - A specific property to dump. If None, all of the object's lazy-loaded properties will be dumped.
        """

        if value:
            val = self._cache.pop(value, None)
            if val and isinstance(val, Object):
                self._client.clear_cache(val)
        else:
            for entry in self._cache:
                val = self._cache[entry]
                if val and isinstance(val, Object):
                    self._client.clear_cache(val)
            self._cache={}

    def create(self):
        """
        Creates a new object in Arena based on object attributes. Not available for objects retrieved from Arena.
        """

        if self.__dict__.get("guid"):
            raise ValueError(f"This {self.__class__.__name__} has already been created")
        data= self._client._post(self.__class__.listing_endpoint, data=self.__dict__)

        self.__dict__.update(data)

    def update(self, data=None):
        """
        Updates the object properties in Arena based on attributes passed in.

        Required arguments:

        - data - A dictionary of object properties to update
        """

        data=data or self.__dict__

        data= self._client._put(self.endpoint, data)

        self.__dict__.update(data)

    def delete(self):
        """
        Deletes the object in Arena.
        """
        data= self._client._delete(self.endpoint)

        self.__dict__.update(data)

    def reload(self):
        """
        Hard reloads the object from Arena. Useful when object is first obtained via a Listing search, which does not include all desired information.
        """

        data = self._client._get(self.endpoint)
        self.__dict__.update(data)

    @property
    def cache_key(self):
        return f"{self.__class__.__name__};{self.guid}"

    @property
    @lazy
    def user(self):
        return self._client.User(self.creator['guid'])
    



from .base import *

class Supplier(Object):
    listing_endpoint="/suppliers"
    endpoint="/suppliers/{guid}"

    @property
    @lazy
    def addresses(self):
        addresses= self._client.Listing(self._client.SupplierAddress, endpoint=f"/suppliers/{self.guid}/addresses")
        for address in addresses:
            address.supplier=self

        return addresses

    @property
    @lazy
    def phones(self):
        phones= self._client.Listing(self._client.SupplierPhone, endpoint=f"/suppliers/{self.guid}/phonenumbers")
        for phone in phones:
            phone.supplier=self

        return phones

    @property
    @lazy
    def quality_processes(self):
        return self._client.Listing(self._client.QualityProcess, endpoint=f"/suppliers/{self.guid}/quality")

    @property
    @lazy
    def files(self):
        files=self._client.Listing(self._client.SupplierFileAssociation, endpoint=f"/suppliers/{self.guid}/files")

        for file in files:
            file.supplier=self

        return files

class SupplierAddress(Object):

    @property
    def listing_endpoint(self):
        return f"/suppliers/{self.supplier.guid}/addresses"
    
    
    @property
    def endpoint(self):
        return f"{self.listing_endpoint}/{self.guid}"

class SupplierFileAssociation(Object):

    @property
    def listing_endpoint(self):
        return f"/suppliers/{self.supplier.guid}/files"

    @property
    def endpoint(self):
        return f"{self.listing_endpoint}/{self.guid}"

    @property
    @lazy
    def file(self):
        return self._client.File(self.__dict__["file"]["guid"])

class SupplierPhone(Object):

    @property
    def listing_endpoint(self):
        return f"/suppliers/{self.supplier.guid}/phonenumbers"
    
    
    @property
    def endpoint(self):
        return f"{self.listing_endpoint}/{self.guid}"


import setuptools

setuptools.setup(
	name="pyrena",
	version="1.0.15",
	description="Python wrapper for Arena QMS API.",
	url="https://github.com/thecodeforge/pyrena",
	py_modules=["pyrena"],
	install_requires=[
		"mistletoe",
		"requests"
	],
	packages=setuptools.find_packages(),
	package_dir={
		"pyrena":"pyrena"
	}
)